import os
from math import log10

EXPORT_RESOLUTION = 300  # dpi
EXPORT_SIZE_X = 19.0  # inch
EXPORT_SIZE_Y = 11.0  # inch
SHOW_PLOTS = False
PLOTTING_DIR = os.path.abspath(os.path.join(os.path.pardir, 'plots'))

NR_TEST_POLYNOMIALS = 5  # repetitions
MAX_COEFF_MAGNITUDE = 1e0
MAX_INP_MAGNITUDE = MAX_COEFF_MAGNITUDE  # max magnitude of evaluation points x
MAX_DIMENSION = 3
DIM_RANGE = list(range(1, MAX_DIMENSION + 1))
MAX_DEGREE = 4
DEGREE_RANGE = list(range(1, MAX_DEGREE + 1))

# speed tests
NR_SAMPLES_SPEED_TEST = 100
SPEED_RUN_PICKLE = 'speed_results.pickle'

# numercial tests
NR_COEFF_CHANGES = 100  # controlling the noise for averaging the numerical error of a single polynomial run
TEST_RESULTS_PICKLE = 'test_results.pickle'

# max allowed numerical error
# n orders of magnitudes less than the coefficients
# maximally the machine precision
MAX_ERR_EXPONENT = max(-15, (int(log10(MAX_COEFF_MAGNITUDE)) - 10))
MAX_NUMERICAL_ERROR = 10 ** MAX_ERR_EXPONENT

# TEST CASES:
INVALID_INPUT_DATA = [
    # calling with x of another dimension
    (([1.0, 2.0, 3.0],
      [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
      [-2.0, 3.0]),
     29.0),

    (([1.0, 2.0, 3.0],
      [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
      [-2.0, 3.0, 1.0, 4.0]),
     29.0),

    # negative exponents are not allowed
    (([1.0, 2.0, 3.0],
      [[3, -1, 0], [2, 0, 1], [1, 1, 1]],
      [-2.0, 3.0, 1.0, 4.0]),
     29.0),

    # duplicate exponent entries are not allowed
    (([1.0, 2.0, 3.0],
      [[3, 1, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]],
      [-2.0, 3.0, 1.0, 4.0]),
     29.0),

]
VALID_TEST_DATA = [
    #
    # p(x) =  5.0
    (([5.0],  # coefficients
      [0],  # exponents
      [0.0]),  # x
     5.0),  # p(x)

    # p(1.0) = 1.0
    (([5.0],
      [0],
      [1.0]),
     5.0),

    # p(-1.0) = -1.0
    (([5.0],
      [0],
      [-1.0]),
     5.0),

    # p(33.5) =33.5
    (([5.0],
      [0],
      [33.5]),
     5.0),

    # p(x) =  1.0* x_1^1
    # p(0.0) = 0.0
    (([1.0],  # coefficients
      [1],  # exponents
      [0.0]),  # x
     0.0),  # p(x)

    # p(1.0) = 1.0
    (([1.0],
      [1],
      [1.0]),
     1.0),

    # p(-1.0) = -1.0
    (([1.0],
      [1],
      [-1.0]),
     -1.0),

    # p(33.5) =33.5
    (([1.0],
      [1],
      [33.5]),
     33.5),

    # p(x) =  1.0* x_1^1 + 1.0 * x_2^1
    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [0.0, 0.0]),
     0.0),

    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [1.0, 0.0]),
     1.0),

    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [-1.0, 0.0]),
     -1.0),

    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [-1.0, 1.0]),
     0.0),

    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [-1.0, -2.0]),
     -3.0),

    (([1.0, 1.0],
      [[1, 0], [0, 1]],
      [33.5, 0.0]),
     33.5),

    # p(x) =  5.0 +  1.0* x_1^1
    (([5.0, 1.0],
      [[0, 0], [1, 0]],
      [0.0, 0.0]),
     5.0),

    (([5.0, 1.0],
      [[0, 0], [1, 0]],
      [1.0, 0.0]),
     6.0),

    (([5.0, 1.0],
      [[0, 0], [1, 0]],
      [-1.0, 0.0]),
     4.0),

    (([5.0, 1.0],
      [[0, 0], [1, 0]],
      [33.5, 0.0]),
     38.5),

    # p(x) =  5.0 + 2.0* x_1^1 + 1.0* x_1^2
    (([5.0, 2.0, 1.0],
      [[0, 0], [1, 0], [2, 0]],
      [0.0, 0.0]),
     5.0),

    (([5.0, 2.0, 1.0],
      [[0, 0], [1, 0], [2, 0]],
      [1.0, 0.0]),
     8.0),  # p(x) =  5.0 + 2.0 + 1.0

    (([5.0, 2.0, 1.0],
      [[0, 0], [1, 0], [2, 0]],
      [-1.0, 0.0]),
     4.0),  # p(x) =  5.0 - 2.0 + 1.0

    (([5.0, 2.0, 1.0],
      [[0, 0], [1, 0], [2, 0]],
      [2.0, 0.0]),
     13.0),  # p(x) =  5.0 + 2.0* 2.0^1 + 1.0* 2.0^2

    # p(x) =  5.0 + 2.0* x_1^1 + 1.0* x_1^2 + 2.0* x_1^2 *x_2^1
    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [0.0, 0.0]),
     5.0),

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [1.0, 0.0]),
     8.0),  # p(x) =  5.0 + 2.0* 1^1 + 1.0* 1^2 + 2.0* 1^2 *0^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [1.0, 1.0]),
     10.0),  # p(x) =  5.0 + 2.0* 1^1 + 1.0* 1^2 + 2.0* 1^2 *1^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [-1.0, 0.0]),
     4.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *0^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [-1.0, 1.0]),
     6.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *1^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [-1.0, 2.0]),
     8.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *2^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [-1.0, 3.0]),
     10.0),  # p(x) =  5.0 + 2.0* (-1)^1 + 1.0* (-1)^2 + 2.0* (-1)^2 *3^1

    (([5.0, 2.0, 1.0, 2.0],
      [[0, 0], [1, 0], [2, 0], [2, 1]],
      [-2.0, 3.0]),
     # p(x) = 5.0 + 2.0* (-2)^1 + 1.0* (-2)^2 + 2.0* (-2)^2 *3^1 = 5.0 + 2.0* (-2) + 1.0* 4 + 2.0* 4 *3
     29.0),

    # as in paper: "Greedy Algorithms for Optimizing Multivariate Horner Schemes"
    # [20] p(x) = 1.0 x_1^3 x_2^1 + 1.0 x_1^2 x_3^1 + 1.0 x_1^2 x_2^1 x_3^1
    # [17] p(x) = x_2^1 [ x_1^3 [ 1.0 ] + x_1^1 x_3^1 [ 1.0 ] ] + x_1^2 x_3^1 [ 1.0 ]
    (([1.0, 1.0, 1.0],
      [[3, 1, 0], [2, 0, 1], [2, 1, 1]],
      [1.0, 1.0, 1.0]),
     3.0),

    # [20] p(x) = 1.0 x_1^3 x_2^1 + 2.0 x_1^2 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1
    # [17] p(x) = x_2^1 [ x_1^3 [ 1.0 ] + x_1^1 x_3^1 [ 3.0 ] ] + x_1^2 x_3^1 [ 2.0 ]
    (([1.0, 2.0, 3.0],
      [[3, 1, 0], [2, 0, 1], [1, 1, 1]],
      [-2.0, 3.0, 1.0]),
     -34.0),

    # [27] p(x) = 1.0 x_3^1 + 2.0 x_1^3 x_2^3 + 3.0 x_1^2 x_2^3 x_3^1 + 4.0 x_1^1 x_2^5 x_3^1
    (([1.0, 2.0, 3.0, 4.0],
      [[0, 0, 1], [3, 3, 0], [2, 3, 1], [1, 5, 1]],
      [-2.0, 3.0, 1.0]),
     -2051.0),
]
COEFF_CHANGE_DATA = [

    # [20] p(x) = 1.0 x_1^3 x_2^1 + 2.0 x_1^2 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1
    # [17] p(x) = x_2^1 [ x_1^3 [ 1.0 ] + x_1^1 x_3^1 [ 3.0 ] ] + x_1^2 x_3^1 [ 2.0 ]
    (([1.0, 1.0, 1.0],  # coeffs1
      [[3, 1, 0], [2, 0, 1], [2, 1, 1]],
      [1.0, 1.0, 1.0],  # coeffs1
      [1.0, 2.0, 3.0],  # coeffs2
      ),
     6.0),
]
